{

******************* uev3WebCam 0.1 *********************

 Файл является частью проекта RubiRobot (http://rubirobot.ru/).

 © 2020-2021 Слинкин Д.А.

 Модуль поддержки видеозахвата с помощью веб-камеры на Linux.
 Для доступа к вебкамере применяется API Video4Linux2.
 Поддерживает захват в формате YUYV.
 Поддерживает экспорт в форматах JPEG, PNG

 Основан на
  fswebcam (https://github.com/fsphil/fswebcam/), GNU GPL v2 License
  5DPO Component Library (https://sourceforge.net/projects/sdpo-cl/), GNU LGPL v2 License
  Video4L2 capture and controls enumeration component for Lazarus (https://github.com/uaply/fpcVideo4L2), MIT License

 Исходный код модуля распространяется под лицензией LGPL v3.0

 ДАННОЕ ПРОГРАММНОЕ ОБЕСПЕЧЕНИЕ ПРЕДОСТАВЛЯЕТСЯ «КАК ЕСТЬ», БЕЗ
 КАКИХ-ЛИБО ГАРАНТИЙ, ЯВНО ВЫРАЖЕННЫХ ИЛИ ПОДРАЗУМЕВАЕМЫХ,
 ВКЛЮЧАЯ ГАРАНТИИ ТОВАРНОЙ ПРИГОДНОСТИ, СООТВЕТСТВИЯ ПО ЕГО
 КОНКРЕТНОМУ НАЗНАЧЕНИЮ И ОТСУТСТВИЯ НАРУШЕНИЙ, НО НЕ ОГРАНИЧИВАЯСЬ
 ИМИ. НИ В КАКОМ СЛУЧАЕ АВТОРЫ ИЛИ ПРАВООБЛАДАТЕЛИ НЕ НЕСУТ
 ОТВЕТСТВЕННОСТИ ПО КАКИМ-ЛИБО ИСКАМ, ЗА УЩЕРБ ИЛИ ПО ИНЫМ ТРЕБОВАНИЯМ,
 В ТОМ ЧИСЛЕ, ПРИ ДЕЙСТВИИ КОНТРАКТА, ДЕЛИКТЕ ИЛИ ИНОЙ СИТУАЦИИ,
 ВОЗНИКШИМ ИЗ-ЗА ИСПОЛЬЗОВАНИЯ ПРОГРАММНОГО ОБЕСПЕЧЕНИЯ ИЛИ ИНЫХ
 ДЕЙСТВИЙ С ПРОГРАММНЫМ ОБЕСПЕЧЕНИЕМ.

}

unit uev3webcam;

{$mode objfpc}{$H+}
{$modeswitch nestedprocvars}

interface

uses cthreads, BaseUnix, fpimage, DateUtils, fpwritejpeg, fpwritepng;


{
   Частичная выборка типов данных и констант API Video for Linux Two
}
(******************************************************************************)
(*******************************videodev2begin*********************************)
(******************************************************************************)

{$PACKRECORDS C}

const
     _IOC_NRBITS = 8;
     _IOC_TYPEBITS = 8;
     _IOC_SIZEBITS = 14;
     _IOC_DIRBITS = 2;
     _IOC_NRMASK = (1 shl _IOC_NRBITS)-1;
     _IOC_TYPEMASK = (1 shl _IOC_TYPEBITS)-1;
     _IOC_SIZEMASK = (1 shl _IOC_SIZEBITS)-1;
     _IOC_DIRMASK = (1 shl _IOC_DIRBITS)-1;
     _IOC_NRSHIFT = 0;
     _IOC_TYPESHIFT = _IOC_NRSHIFT+_IOC_NRBITS;
     _IOC_SIZESHIFT = _IOC_TYPESHIFT+_IOC_TYPEBITS;
     _IOC_DIRSHIFT = _IOC_SIZESHIFT+_IOC_SIZEBITS;
     _IOC_NONE = 0;
     _IOC_WRITE = 1;
     _IOC_READ = 2;

type
  __u8 = byte;
  __s8 = shortint;
  __u16 = word;
  __s16 = smallint;
  __u32 = LongWord;
  __s32 = LongInt;
  __u64 = QWord;
  __s64 = Int64;

  const
    V4L2_CTRL_CLASS_USER         = $00980000; { Old-style 'user' controls  }
    V4L2_CID_BASE = V4L2_CTRL_CLASS_USER or $900;    
    V4L2_CID_USER_BASE = V4L2_CID_BASE;    
    V4L2_CID_USER_CLASS = V4L2_CTRL_CLASS_USER or 1;    
    V4L2_CID_BRIGHTNESS = V4L2_CID_BASE+0;    
    V4L2_CID_CONTRAST = V4L2_CID_BASE+1;    
    V4L2_CID_SATURATION = V4L2_CID_BASE+2;    
    V4L2_CID_HUE = V4L2_CID_BASE+3;    
    V4L2_CID_AUTO_WHITE_BALANCE = V4L2_CID_BASE+12;    
    V4L2_CID_DO_WHITE_BALANCE = V4L2_CID_BASE+13;    
    V4L2_CID_RED_BALANCE = V4L2_CID_BASE+14;    
    V4L2_CID_BLUE_BALANCE = V4L2_CID_BASE+15;    
    V4L2_CID_GAMMA = V4L2_CID_BASE+16;
    V4L2_CID_WHITENESS = V4L2_CID_GAMMA; { Deprecated  }
    V4L2_CID_EXPOSURE = V4L2_CID_BASE+17;    
    V4L2_CID_AUTOGAIN = V4L2_CID_BASE+18;    
    V4L2_CID_GAIN = V4L2_CID_BASE+19;    
    V4L2_CID_HFLIP = V4L2_CID_BASE+20;    
    V4L2_CID_VFLIP = V4L2_CID_BASE+21;    
    V4L2_CID_POWER_LINE_FREQUENCY = V4L2_CID_BASE+24;    
    V4L2_CID_HUE_AUTO = V4L2_CID_BASE+25;    
    V4L2_CID_WHITE_BALANCE_TEMPERATURE = V4L2_CID_BASE+26;    
    V4L2_CID_SHARPNESS = V4L2_CID_BASE+27;    
    V4L2_CID_BACKLIGHT_COMPENSATION = V4L2_CID_BASE+28;    
    V4L2_CID_CHROMA_AGC = V4L2_CID_BASE+29;    
    V4L2_CID_COLOR_KILLER = V4L2_CID_BASE+30;    
    V4L2_CID_COLORFX = V4L2_CID_BASE+31;    
    V4L2_CID_LASTP1 = V4L2_CID_BASE+43;    

  const
    VIDEO_MAX_FRAME = 32;    
    VIDEO_MAX_PLANES = 8;    

  const
      V4L2_FIELD_ANY = 0;
      V4L2_BUF_TYPE_VIDEO_CAPTURE = 1;
      V4L2_MEMORY_MMAP = 1;
      V4L2_MEMORY_USERPTR = 2;
      V4L2_MEMORY_OVERLAY = 3;
      V4L2_MEMORY_DMABUF = 4;

  type
    Pv4l2_rect = ^v4l2_rect;
    v4l2_rect = record
        left : __s32;
        top : __s32;
        width : __s32;
        height : __s32;
      end;

    Pv4l2_fract = ^v4l2_fract;
    v4l2_fract = record
        numerator : __u32;
        denominator : __u32;
      end;



    Pv4l2_capability = ^v4l2_capability;
    v4l2_capability = record
        driver : array[0..15] of __u8;
        card : array[0..31] of __u8;
        bus_info : array[0..31] of __u8;
        version : __u32;
        capabilities : __u32;
        device_caps : __u32;
        reserved : array[0..2] of __u32;
      end;


  type
    Pv4l2_pix_format = ^v4l2_pix_format;
    v4l2_pix_format = record
        width : __u32;
        height : __u32;
        pixelformat : __u32;
        field : __u32;
        bytesperline : __u32;
        sizeimage : __u32;
        colorspace : __u32;
        priv : __u32;
      end;

 const
  V4L2_PIX_FMT_YUYV    = (Ord('Y')) or (Ord('U') shl 8) or (Ord('Y') shl 16) or (Ord('V') shl 24); { 16  YUV 4:2:2     *}


  type
    Pv4l2_fmtdesc = ^v4l2_fmtdesc;
    v4l2_fmtdesc = record
        index : __u32;
        _type : __u32;
        flags : __u32;
        description : array[0..31] of __u8;
        pixelformat : __u32;
        reserved : array[0..3] of __u32;
      end;

  type
    v4l2_frmsizetypes =  Longint;
    Const
      V4L2_FRMSIZE_TYPE_DISCRETE = 1;
      V4L2_FRMSIZE_TYPE_CONTINUOUS = 2;
      V4L2_FRMSIZE_TYPE_STEPWISE = 3;

  type
    Pv4l2_frmsize_discrete = ^v4l2_frmsize_discrete;
    v4l2_frmsize_discrete = record
        width : __u32;
        height : __u32;
      end;

    Pv4l2_frmsize_stepwise = ^v4l2_frmsize_stepwise;
    v4l2_frmsize_stepwise = record
        min_width : __u32;
        max_width : __u32;
        step_width : __u32;
        min_height : __u32;
        max_height : __u32;
        step_height : __u32;
      end;

    Pv4l2_frmsizeenum = ^v4l2_frmsizeenum;
    v4l2_frmsizeenum = record
        index : __u32;
        pixel_format : __u32;
        _type : __u32;
        u : record
            case longint of
              0 : ( discrete : v4l2_frmsize_discrete );
              1 : ( stepwise : v4l2_frmsize_stepwise );
            end;
        reserved : array[0..1] of __u32;
      end;

  type
    Pv4l2_frmival_stepwise = ^v4l2_frmival_stepwise;
    v4l2_frmival_stepwise = record
        min : v4l2_fract;
        max : v4l2_fract;
        step : v4l2_fract;
      end;

    Pv4l2_frmivalenum = ^v4l2_frmivalenum;
    v4l2_frmivalenum = record
        index : __u32;
        pixel_format : __u32;
        width : __u32;
        height : __u32;
        _type : __u32;
        u : record
            case longint of
              0 : ( discrete : v4l2_fract );
              1 : ( stepwise : v4l2_frmival_stepwise );
            end;
        reserved : array[0..1] of __u32;
      end;

  type
    Pv4l2_timecode = ^v4l2_timecode;
    v4l2_timecode = record
        _type : __u32;
        flags : __u32;
        frames : __u8;
        seconds : __u8;
        minutes : __u8;
        hours : __u8;
        userbits : array[0..3] of __u8;
      end;

  type
    Pv4l2_requestbuffers = ^v4l2_requestbuffers;
    v4l2_requestbuffers = record
        count : __u32;
        _type : __u32;
        memory : __u32;
        reserved : array[0..1] of __u32;
      end;

    Pv4l2_plane = ^v4l2_plane;
    v4l2_plane = record
        bytesused : __u32;
        length : __u32;
        m : record
            case longint of
              0 : ( mem_offset : __u32 );
              {$ifdef CPU32}
              1 : ( userptr : dword );
              {$ENDIF}
              {$ifdef CPU64}
              1 : ( userptr : __u64 );
              {$ENDIF}
              2 : ( fd : __s32 );
            end;
        data_offset : __u32;
        reserved : array[0..10] of __u32;
      end;
    Pv4l2_buffer = ^v4l2_buffer;
    v4l2_buffer = record
        index : __u32;
        _type : __u32;
        bytesused : __u32;
        flags : __u32;
        field : __u32;
        timestamp : timeval;
        timecode : v4l2_timecode;
        sequence : __u32;
        memory : __u32;
        m : record
            case longint of
              0 : ( offset : __u32 );
              {$ifdef CPU32}
              1 : ( userptr : dword );
              {$ENDIF}
              {$ifdef CPU64}
              1 : ( userptr : __u64 );
              {$ENDIF}
              2 : ( planes : Pv4l2_plane );
              3 : ( fd : __s32 );
            end;
        length : __u32;
        reserved2 : __u32;
        reserved : __u32;
      end;
  type
    Pv4l2_exportbuffer = ^v4l2_exportbuffer;
    v4l2_exportbuffer = record
        _type : __u32;
        index : __u32;
        plane : __u32;
        flags : __u32;
        fd : __s32;
        reserved : array[0..10] of __u32;
      end;
  type
    Pv4l2_clip = ^v4l2_clip;
    v4l2_clip = record
        c : v4l2_rect;
        next : Pv4l2_clip;
      end;

    Pv4l2_window = ^v4l2_window;
    v4l2_window = record
        w : v4l2_rect;
        field : __u32;
        chromakey : __u32;
        clips : Pv4l2_clip;
        clipcount : __u32;
        bitmap : pointer;
        global_alpha : __u8;
      end;

    Pv4l2_captureparm = ^v4l2_captureparm;
    v4l2_captureparm = record
        capability : __u32;
        capturemode : __u32;
        timeperframe : v4l2_fract;
        extendedmode : __u32;
        readbuffers : __u32;
        reserved : array[0..3] of __u32;
      end;

    Pv4l2_outputparm = ^v4l2_outputparm;
    v4l2_outputparm = record
        capability : __u32;
        outputmode : __u32;
        timeperframe : v4l2_fract;
        extendedmode : __u32;
        writebuffers : __u32;
        reserved : array[0..3] of __u32;
      end;

    Pv4l2_control = ^v4l2_control;
    v4l2_control = record
        id : __u32;
        value : __s32;
      end;

    Pv4l2_queryctrl = ^v4l2_queryctrl;
    v4l2_queryctrl = record
        id : __u32;
        _type : __u32;
        name : array[0..31] of char; // of __u8;
        minimum : __s32;
        maximum : __s32;
        step : __s32;
        default_value : __s32;
        flags : __u32;
        reserved : array[0..1] of __u32;
      end;

  const
    V4L2_CTRL_FLAG_DISABLED = $0001;    
    V4L2_CTRL_FLAG_GRABBED = $0002;    
    V4L2_CTRL_FLAG_READ_ONLY = $0004;    
    V4L2_CTRL_FLAG_UPDATE = $0008;    
    V4L2_CTRL_FLAG_INACTIVE = $0010;    
    V4L2_CTRL_FLAG_SLIDER = $0020;    
    V4L2_CTRL_FLAG_WRITE_ONLY = $0040;    
    V4L2_CTRL_FLAG_VOLATILE = $0080;    
    V4L2_CTRL_FLAG_NEXT_CTRL = $80000000;    
    V4L2_CID_MAX_CTRLS = 1024;    
    V4L2_CID_PRIVATE_BASE = $08000000;    

  type
    Pv4l2_vbi_format = ^v4l2_vbi_format;
    v4l2_vbi_format = record
        sampling_rate : __u32;
        offset : __u32;
        samples_per_line : __u32;
        sample_format : __u32;
        start : array[0..1] of __s32;
        count : array[0..1] of __u32;
        flags : __u32;
        reserved : array[0..1] of __u32;
      end;

    Pv4l2_sliced_vbi_format = ^v4l2_sliced_vbi_format;
    v4l2_sliced_vbi_format = record
        service_set : __u16;
        service_lines : array[0..1] of array[0..23] of __u16;
        io_size : __u32;
        reserved : array[0..1] of __u32;
      end;

    v4l2_plane_pix_format = packed record
        sizeimage : __u32;
        bytesperline : __u16;
        reserved : array[0..6] of __u16;
      end;

    Pv4l2_pix_format_mplane = ^v4l2_pix_format_mplane;
    v4l2_pix_format_mplane = packed record
        width : __u32;
        height : __u32;
        pixelformat : __u32;
        field : __u32;
        colorspace : __u32;
        plane_fmt : array[0..(VIDEO_MAX_PLANES)-1] of v4l2_plane_pix_format;
        num_planes : __u8;
        reserved : array[0..10] of __u8;
      end;


    Pv4l2_format = ^v4l2_format;
    v4l2_format = record
        _type : __u32;
        fmt : record
            case longint of
              0 : ( pix : v4l2_pix_format );
              1 : ( pix_mp : v4l2_pix_format_mplane );
              2 : ( win : v4l2_window );
              3 : ( vbi : v4l2_vbi_format );
              4 : ( sliced : v4l2_sliced_vbi_format );
              5 : ( raw_data : array[0..199] of __u8 );
            end;
      end;

    Pv4l2_streamparm = ^v4l2_streamparm;
    v4l2_streamparm = record
        _type : __u32;
        parm : record
            case longint of
              0 : ( capture : v4l2_captureparm );
              1 : ( output : v4l2_outputparm );
              2 : ( raw_data : array[0..199] of __u8 );
            end;
      end;
const
  VIDIOC_QUERYCAP          = LongInt((_IOC_READ shl _IOC_DIRSHIFT) or (Ord('V') shl _IOC_TYPESHIFT) or (0 shl _IOC_NRSHIFT) or (SizeOf(v4l2_capability) shl _IOC_SIZESHIFT));
  VIDIOC_ENUM_FMT         = LongInt(((_IOC_READ or _IOC_WRITE) shl _IOC_DIRSHIFT) or (Ord('V') shl _IOC_TYPESHIFT) or (2 shl _IOC_NRSHIFT) or (SizeOf(v4l2_fmtdesc) shl _IOC_SIZESHIFT));

  VIDIOC_G_FMT            = LongInt(((_IOC_READ or _IOC_WRITE) shl _IOC_DIRSHIFT) or (Ord('V') shl _IOC_TYPESHIFT) or (4 shl _IOC_NRSHIFT) or (SizeOf(v4l2_format) shl _IOC_SIZESHIFT));
  VIDIOC_S_FMT            = LongInt(((_IOC_READ or _IOC_WRITE) shl _IOC_DIRSHIFT) or (Ord('V') shl _IOC_TYPESHIFT) or (5 shl _IOC_NRSHIFT) or (SizeOf(v4l2_format) shl _IOC_SIZESHIFT));
  VIDIOC_REQBUFS          = LongInt(((_IOC_READ or _IOC_WRITE) shl _IOC_DIRSHIFT) or (Ord('V') shl _IOC_TYPESHIFT) or (8 shl _IOC_NRSHIFT) or (SizeOf(v4l2_requestbuffers) shl _IOC_SIZESHIFT));
  VIDIOC_QUERYBUF         = LongInt(((_IOC_READ or _IOC_WRITE) shl _IOC_DIRSHIFT) or (Ord('V') shl _IOC_TYPESHIFT) or (9 shl _IOC_NRSHIFT) or (SizeOf(v4l2_buffer) shl _IOC_SIZESHIFT));
  VIDIOC_QBUF             = LongInt(((_IOC_READ or _IOC_WRITE) shl _IOC_DIRSHIFT) or (Ord('V') shl _IOC_TYPESHIFT) or (15 shl _IOC_NRSHIFT) or (SizeOf(v4l2_buffer) shl _IOC_SIZESHIFT));
  VIDIOC_EXPBUF           = LongInt(((_IOC_READ or _IOC_WRITE) shl _IOC_DIRSHIFT) or (Ord('V') shl _IOC_TYPESHIFT) or (16 shl _IOC_NRSHIFT) or (SizeOf(v4l2_exportbuffer) shl _IOC_SIZESHIFT));
  VIDIOC_DQBUF            = LongInt(((_IOC_READ or _IOC_WRITE) shl _IOC_DIRSHIFT) or (Ord('V') shl _IOC_TYPESHIFT) or (17 shl _IOC_NRSHIFT) or (SizeOf(v4l2_buffer) shl _IOC_SIZESHIFT));

  VIDIOC_STREAMON          = LongInt((_IOC_WRITE shl _IOC_DIRSHIFT) or (Ord('V') shl _IOC_TYPESHIFT) or (18 shl _IOC_NRSHIFT) or (SizeOf(LongInt) shl _IOC_SIZESHIFT));
  VIDIOC_STREAMOFF         = LongInt((_IOC_WRITE shl _IOC_DIRSHIFT) or (Ord('V') shl _IOC_TYPESHIFT) or (19 shl _IOC_NRSHIFT) or (SizeOf(LongInt) shl _IOC_SIZESHIFT));
  VIDIOC_S_PARM           = LongInt(((_IOC_READ or _IOC_WRITE) shl _IOC_DIRSHIFT) or (Ord('V') shl _IOC_TYPESHIFT) or (22 shl _IOC_NRSHIFT) or (SizeOf(v4l2_streamparm) shl _IOC_SIZESHIFT));
  VIDIOC_G_CTRL           = LongInt(((_IOC_READ or _IOC_WRITE) shl _IOC_DIRSHIFT) or (Ord('V') shl _IOC_TYPESHIFT) or (27 shl _IOC_NRSHIFT) or (SizeOf(v4l2_control) shl _IOC_SIZESHIFT));
  VIDIOC_S_CTRL           = LongInt(((_IOC_READ or _IOC_WRITE) shl _IOC_DIRSHIFT) or (Ord('V') shl _IOC_TYPESHIFT) or (28 shl _IOC_NRSHIFT) or (SizeOf(v4l2_control) shl _IOC_SIZESHIFT));
  VIDIOC_QUERYCTRL        = LongInt(((_IOC_READ or _IOC_WRITE) shl _IOC_DIRSHIFT) or (Ord('V') shl _IOC_TYPESHIFT) or (36 shl _IOC_NRSHIFT) or (SizeOf(v4l2_queryctrl) shl _IOC_SIZESHIFT));
  VIDIOC_ENUM_FRAMESIZES  = LongInt(((_IOC_READ or _IOC_WRITE) shl _IOC_DIRSHIFT) or (Ord('V') shl _IOC_TYPESHIFT) or (74 shl _IOC_NRSHIFT) or (SizeOf(v4l2_frmsizeenum) shl _IOC_SIZESHIFT));
  VIDIOC_ENUM_FRAMEINTERVALS = LongInt(((_IOC_READ or _IOC_WRITE) shl _IOC_DIRSHIFT) or (Ord('V') shl _IOC_TYPESHIFT) or (75 shl _IOC_NRSHIFT) or (SizeOf(v4l2_frmivalenum) shl _IOC_SIZESHIFT));

(******************************************************************************)
(*********************************videodev2end*********************************)
(******************************************************************************)



type
  // YUYV - поддерживаемый подавляющим большинством камер формат данных
  // С учетом специфики применения проекта RubiRobot - вполне достаточно

  TPixelFormat=(pf_NONE,pf_YUYV);

const
  ConstsUVCPixelFormat: array[TPixelFormat] of longint = (0,V4L2_PIX_FMT_YUYV);

const
    ParamMax=high(int32);
    ParamMin=low(int32);
    ParamNone=-1;

type
  TWebCam = class;
  TWebCamFrameEvent = function(Sender: TWebCam; FramePtr: PByte; FrameNumber:integer):boolean of object;
  TWebCamFrameEventProc = function(Sender: TWebCam; FramePtr: PByte; FrameNumber:integer):boolean;
  TWebCamFrameEventNestedProc =
      function(Sender: TWebCam; FramePtr: PByte; FrameNumber:integer):boolean is nested;
  TUserControl=class;
  TArrObj=class;

  { TV4L2Control }

  TV4L2Control=class
   private
    fowner:TWebCam;
   protected
    property owner:TWebCam read fowner;
   public
    constructor Create(_owner:TWebCam);
    function valid:boolean;
  end;

  { TFrameRate }

  TFrameRate=class(TV4L2Control)
   private
    fframerate:v4l2_frmivalenum;
   public
    constructor Create(_owner:TWebCam; framerate:v4l2_frmivalenum);
    property width:dword read fframerate.width;
    property height:dword read fframerate.height;
    property numerator:dword read fframerate.u.discrete.numerator;
    property denominator:dword read fframerate.u.discrete.denominator;
    property rate:dword read fframerate.u.discrete.denominator;
    procedure setFrameRate();
  end;

  { TPixelFormatDescription }

  TPixelFormatDescription=class(TV4L2Control)
    private
     ffmt:v4l2_fmtdesc;
     function getDescription: string;
     function getpixelFormat: TPixelFormat;
     function getpixelFormatName: string;
    public
     constructor Create(_owner:TWebCam; fmt:v4l2_fmtdesc);
    public
     property description:string read getDescription;
     property pixelFormat:TPixelFormat read getpixelFormat;
     property Name:string read getpixelFormatName;
     property V4L2_pixelFormat:dword read ffmt.pixelformat;
  end;


  { TFrameControl }

  TFrameControl=class(TV4L2Control)
    private
     fframe:v4l2_frmsizeenum;
     fframerates:TArrObj;
     function getFrameRate(index: integer): TFrameRate;
     function getFrameRateCount: integer;
    public
      constructor Create(_owner:TWebCam; frame:v4l2_frmsizeenum);
      destructor Destroy;override;
    public
     property frameRateCount:integer read getFrameRateCount;
     property frameRate[index:integer]:TFrameRate read getFrameRate; default;
     property width:dword read fframe.u.discrete.width;
     property height:dword read fframe.u.discrete.height;
     procedure setFrameRate(Rate:double=ParamMax);
  end;

  { TUserControl }

  TUserControl=class(TV4L2Control)
    private
      fcontrol:v4l2_queryctrl;
      function getName: string;
      function getValue: integer;
      procedure setValue(AValue: integer);
    public
      constructor Create(_owner:TWebCam; control:v4l2_queryctrl);
      property id:dword read fcontrol.id;
      property name:string read getName;
      property min:integer read fcontrol.minimum;
      property max:integer read fcontrol.maximum;
      property step:integer read fcontrol.step;
      property def:integer read fcontrol.default_value;
      property value:integer read getValue write setValue;
      procedure setDefault;
      procedure setMin;
      procedure setMax;
  end;


  PColorRGB=^TColorRGB;
  TColorRGB=TFPCompactImgRGB8BitValue;

  PColorBGRA=^TColorBGRA;
  TColorBGRA=packed record
    b,g,r,a:byte;
  end;

  PColorGray=^TColorGray;
  TColorGray=byte;

  { TCamImage }

  TCamImage=class(TFPCompactImgBase)
    private
      FData: PByte;
      FDesc: TFPCompactImgDesc;
    protected
      function GetInternalPixel({%H-}x, {%H-}y: integer): integer; override;
      procedure SetInternalPixel({%H-}x, {%H-}y: integer; {%H-}Value: integer); override;
    public
      property Desc: TFPCompactImgDesc read FDesc;
      property Data:PByte read FData;
      destructor Destroy; override;
    public
      function YUYVImport(YUYV:PByte; _Width, _Height : integer):TCamImage; virtual; abstract;
  end;


  { TCamImageRGB }

  TCamImageRGB=class(TCamImage)
   protected
    function GetInternalColor(x, y: integer): TFPColor; override;
    procedure SetInternalColor (x, y: integer; const Value: TFPColor); override;
   public
    constructor Create(AWidth, AHeight: integer); override;
    procedure SetSize(AWidth, AHeight: integer); override;
   public
    function YUYVImport(YUYV:PByte; _Width, _Height : integer):TCamImage;override;
  end;


  { TCamImageBGRA }

  TCamImageBGRA=class(TCamImage)
   protected
    function GetInternalColor(x, y: integer): TFPColor; override;
    procedure SetInternalColor (x, y: integer; const Value: TFPColor); override;
   public
    constructor Create(AWidth, AHeight: integer); override;
    procedure SetSize(AWidth, AHeight: integer); override;
   public
    function YUYVImport(YUYV:PByte; _Width, _Height : integer):TCamImage;override;
  end;


  { TCamImageGray }

  TCamImageGray=class(TCamImage)
   protected
    function GetInternalColor(x, y: integer): TFPColor; override;
    procedure SetInternalColor (x, y: integer; const Value: TFPColor); override;
   public
    constructor Create(AWidth, AHeight: integer); override;
    procedure SetSize(AWidth, AHeight: integer); override;
   public
    function YUYVImport(YUYV:PByte; _Width, _Height : integer):TCamImage;override;
  end;


  { TWebCam }

  type
   // возможности веб-камеры
   TCamCap=(
	CAP_VIDEO_CAPTURE,
	CAP_VIDEO_OUTPUT,
	CAP_VIDEO_OVERLAY,
        CAP_DUMMY_3,
	CAP_VBI_CAPTURE,
	CAP_VBI_OUTPUT,
	CAP_SLICED_VBI_CAPTURE,
	CAP_SLICED_VBI_OUTPUT,
	CAP_RDS_CAPTURE,
	CAP_VIDEO_OUTPUT_OVERLAY,
	CAP_HW_FREQ_SEEK,
	CAP_RDS_OUTPUT,
        CAP_VIDEO_CAPTURE_MPLANE,
        CAP_VIDEO_OUTPUT_MPLANE,
        CAP_VIDEO_M2M,
        CAP_VIDEO_M2M_MPLANE,
        CAP_TUNER,
        CAP_AUDIO,
        CAP_RADIO,
        CAP_MODULATOR,
        CAP_SDR_CAPTURE,
        CAP_EXT_PIX_FORMAT,
        CAP_SDR_OUTPUT,
        CAP_META_CAPTURE,
        CAP_READWRITE,
        CAP_ASYNCIO,
        CAP_STREAMING,
        CAP_CAP_META_OUTPUT,
        CAP_TOUCH,
        CAP_DUMMY_29,
        CAP_DUMMY_30,
        CAP_DEVICE_CAPS
    );
   TCamCaps=set of TCamCap;

 type
  TWebCam=class
  private
   const VIDEO_BUFFERS = 4;
   type  TV4L2_CIDS = V4L2_CID_BASE..V4L2_CID_LASTP1;
  private
    fCap:v4l2_capability;
    fDriver,fCard,fBusInfo:string;
    fCamCaps:TCamCaps;

    fHeight: integer;
    fPixelFormat: TPixelFormat;
    fRateNum,fRateDenom: integer;
    fWidth: integer;

    FOnFrame: TWebCamFrameEvent;

    FOpened: boolean;
    FHandle: integer;
    FPrefixDevice: string;
    FDevice: string;
    FStarted: boolean;

    VideoBuffer: array[0..VIDEO_BUFFERS - 1] of pbyte;
    VideoBufferLength: integer;
    VideoBufferIndex: integer;

    function CapabilitiesToCamCaps(v4l2caps:__u32):TCamCaps;inline;
    function getCamCapInfo(index: TCamCap): ansistring;
    function GetFrameControl(index: integer): TFrameControl;
    function GetFramesCount: integer;
    function getV4L2_PixelFormat: longint;

    procedure getParams(var _Width:integer; var _Height:integer; _Rate:double; out _RateNum:integer; out _RateDenom:integer);
    function getRate: double;

    procedure GetUserControls;
    procedure GetFrameControls;
    procedure GetPixelFormats;

    function GetUserControl(index: integer): TUserControl;
    function GetUserControlsCount: integer;
    function GetExtControl(index: integer): TUserControl;
    function GetExtControlsCount: integer;

  protected
    property Handle: integer read FHandle;
    property V4L2_PixelFormat: longint read getV4L2_PixelFormat;

    procedure ClearDebugList();
    procedure AddDebug(s: string);

    function GetFeatureValue(feature: LongWord): integer;
    procedure SetFeatureValue(feature: LongWord; value: integer);
    function TrySetFeatureValue(feature: LongWord; value: integer): boolean;

  protected
    // базовый набор параметров камеры
    _UserControls: array[TV4L2_CIDS] of TUserControl;
    // дополнительный набор параметров камеры
    _ExtControls: TArrObj;
    // доступные разрешения и скорости съемки камеры
    _Frames: TArrObj;
    // доступные форматы данных камеры
    _PixelFormats: TArrObj;

  public
    const PAL_YUYV = 8;  PAL_UYVY = 9; PAL_VYUY = 10;

    // Преобразователи форматов изображений.
    // Размеры исходного и целевого изображений должны быть идентичны
    // При экспорте в TBitmap следует также применять TBitmap.BeginUpdate, TBitmap.EndUpdate

    // src: изображение в YUYV-формате, например - TWebCam.ImageYUYV
    // dest: RGB-триплет в трех байтах на точку, например - TCamImageRGB.Data
    class function dec_yuyv_rgb(src:pbyte; dest:pbyte; width,height:integer; pal:integer=PAL_YUYV):boolean;
    // src: изображение в YUYV-формате, например - TWebCam.ImageYUYV
    // dest: BGRA-квартет в четырех байтах на точку, например - TCamImageBGRA.Data, TBitmap.RawImage.Data
    class function dec_yuyv_bgra(src:pbyte; dest:pbyte; width,height:integer; pal:integer=PAL_YUYV):boolean;
    // src: изображение в YUYV-формате, например - TWebCam.ImageYUYV
    // dest: интенсивность серого в одном байте на точку, например TCamImageGray.Data
    class function dec_yuyv_gray(src:pbyte; dest:pbyte; width,height:integer):boolean;
    // src: изображение в YUYV-формате, например - TWebCam.ImageYUYV
    // dest: 0 или 255 в одном байте на точку, например TCamImageGray.Data
    class function dec_yuyv_bw(src:pbyte; dest:pbyte; width,height:integer; bound:byte):boolean;

    // src: изображение в YUYV-формате, например - TWebCam.ImageYUYV
    // dest: 0 или 255 в одном байте на точку, например TCamImageGray.Data
    // динамический тресхолд-анализ
    // потоконебезопасная функция, использует глобальный массив для хранения промежуточных результатов.
    class function dec_yuyv_bwdynamic(src:pbyte; dest:pbyte; width,height:integer; divX,divY,c:integer):boolean;

    // src: изображение в оттенках серого 1 байт на точку, например TCamImageGray.Data
    // dest: BGRA-квартет в четырех байтах на точку, например - TCamImageBGRA.Data, TBitmap.RawImage.Data
    class function dec_gray_bgra(src:pbyte; dest:pbyte; width,height:integer):boolean;

  public

    // факт подключения к камере
    property Opened: boolean read FOpened;
    // факт готовности к видеозахвату
    property Started: boolean read FStarted;

    // Создает объект вебкамеры. Готовит струтуры данных для работы с вебкамерой.
    // НЕ выполняет реального подключения, для этого следует вызвать DeviceOpen
    constructor Create(_prefixDevice:string='/dev/video');
    // Уничтожает объект вебкамеры, освобождает память.
    destructor Destroy; override;

    {
     Первое действие после создания объекта.

     Подключается к выбранной камере _Device, загружает ее характеристики,
     готовит к установке заявленное разрешение, скорость съемки, формат данных.
     Если камеры с таким номером не существует, пытается найти следующую камеру.
     Если камеры не найдено, генерирует исключение.
     Устанавливает разрешение камеры (_Width*_Height). Если камера не поддерживает
     выбранного разрешения, устанавливает разрешение, рекомендуемое системой.
     Устанавливает скорость съемки (_Rate). Если камера не поддерживает
     выбранной частоты, устанавливает частоту, рекомендуемую системой для
     выбраного разрешения.

     Значение параметра ParamMin означает, что следует выбрать минимально
     возможное значение параметра.
     Значение параметра ParamMax означает, что следует выбрать максимально
     возможное значение параметра.
     Значение параметра ParamNone означает, что следует использовать
     ранее установленное значение параметра. Это возможно при повторном
     подключении к камере.
     Любое другое значение параметра означает использование его без измененеий.
    }

    function DeviceOpen(
                         _Device:integer = ParamMin; // номер камеры, начиная с 0
                         _Width:integer = ParamMax;  // ширина
                         _Height:integer = ParamMax; // высота
                         _Rate:double = ParamMax; // количество кадров в секунду
                         _pixelFormat:TPixelFormat=pf_YUYV // формат данных
                         ):boolean;

    // Устанавливает разрешение и скорость съемки для подключенной камеры.
    // Если камера отключена, то подключает ее с параметрами по умолчанию.
    // Обычно используется при необходимости изменения
    // характеристик захвата после вызова DeviceOpen
    procedure setFrameRate(_Width:integer = ParamMax;  // ширина
                           _Height:integer = ParamMax; // высота
                           _Rate:double = ParamMax // количество кадров в секунду
                          );


    // Отключается от камеры
    // Вызывается автоматически при переподключении к камере или
    // завершении работы с камерой
    procedure DeviceClose;

    // Применяет все ранее установленные параметры, готовит камеру к видеозахвату
    // Вызывается автоматически при старте захвата (capture-функции)
    function DeviceStart:boolean;

    // завершает видеозахват, очищает память
    // вызывается автоматически при изменении параметров камеры после завхвата или
    // завершении работы с камерой
    procedure DeviceStop;

  public
    // информация о последней возникшей ошибке
    LastError:string;
    // информация о всех параметрах камеры в человекочитаемом формате
    DebugList: array of AnsiString;
    // инофрмация о камере
    // используемый драйвер
    property Driver:string read fDriver;
    // название камеры
    property Card:string read fCard;
    // точка подключения камеры
    property BusInfo:string read fBusInfo;
    // возможности камеры в текущем режиме
    property CamCaps:TCamCaps read fCamCaps;
    // текстовое описание возможностей камеры в текущем режиме
    property CamCapInfo[index:TCamCap]:ansistring read getCamCapInfo;

    // часто используемые параметры камеры
    property Brightness: TUserControl read _UserControls[V4L2_CID_BRIGHTNESS];
    property Contrast: TUserControl read _UserControls[V4L2_CID_CONTRAST];
    property Saturation: TUserControl read _UserControls[V4L2_CID_SATURATION];
    property Hue: TUserControl read _UserControls[V4L2_CID_HUE];
    property Gamma: TUserControl read _UserControls[V4L2_CID_GAMMA];
    property Sharpness: TUserControl read _UserControls[V4L2_CID_SHARPNESS];
    property BacklightCompensation: TUserControl read _UserControls[V4L2_CID_BACKLIGHT_COMPENSATION];

    // устанавливает часто используемые параметры камеры в значение по умолчанию
    procedure setDefault;

    // базовый набор параметров камеры, включает в себя часто используемые параметры
    property UserControls[index:integer]:TUserControl read GetUserControl;
    property UserControlsCount:integer read GetUserControlsCount;

    // дополнительный набор параметров камеры
    property ExtControls[index:integer]:TUserControl read GetExtControl;
    property ExtControlsCount:integer read GetExtControlsCount;

    // доступные разрешения и скорости съемки камеры
    property Frames[index:integer]:TFrameControl read GetFrameControl;
    property FramesCount:integer read GetFramesCount;

  public
    // установленные параметры камеры
    property Device: string read FDevice;
    property Width:integer read fWidth;
    property Height:integer read fHeight;
    property RateNum:integer read fRateNum;
    property RateDenum:integer read fRateDenom;
    property Rate:double read getRate; // RateDenum/RateNum - кол-во кадров в секунду
    property PixelFormat: TPixelFormat read fPixelFormat; // в текущей версии только YUYV-формат

  public
    // номер текущего кадра, захватываемого с помощью Capture
    CurCaptureFrame:integer;
    // текущий захваченный кадр, валиден исключительно внутри Capture
    CurFramePtr:pbyte;

    // обработчик захвата единичного кадра
    property OnFrame: TWebCamFrameEvent read FOnFrame write FOnFrame;
    // процедуры старта видеозахвата, count - количество кадров захвата
    // захват с вызовом onFrame для каждого кадра
    function Capture(count:integer=1):boolean;
    // захват с вызовом proc для каждого кадра
    function Capture(proc:TWebCamFrameEventProc; count:integer=1):boolean;
    function Capture(proc:TWebCamFrameEventNestedProc; count:integer=1):boolean;


    // захват одного изображения с предварительным пропуском skip кадров
    // захват RGB-изображения, доступного в дальнейшем через свойство ImageRGB
    function CaptureImageRGB(skip:integer=0):TCamImageRGB;
    // захват BGRA-изображения, доступного в дальнейшем через свойство ImageBGRA
    function CaptureImageBGRA(skip:integer=0):TCamImageBGRA;
    // захват изображения в оттенках серого, доступного в дальнейшем через свойство ImageGray
    function CaptureImageGray(skip:integer=0):TCamImageGray;

    // захват изображения в формате YUYV, доступного в дальнейшем через свойство ImageYUYV
    function CaptureImageYUYV(skip:integer=0):PByte;

  private
    fImageRGB:TCamImageRGB;
    fImageBGRA:TCamImageBGRA;
    fImageGray:TCamImageGray;
    fImageYUYV:PByte;
    fImageYUYVWidth,fImageYUYVHeight:integer;
  public
    // результаты захватов изображений, действительны до следующих захватов
    // или изменений параметров камеры.
    property ImageRGB:TCamImageRGB read fImageRGB;
    property ImageBGRA:TCamImageBGRA read fImageBGRA;
    property ImageGray:TCamImageGray read fImageGray;
    property ImageYUYV:PByte read fImageYUYV;
 end;


// Возвращает созданный при старте программы объект вебкамеры.
function webcam:TWebCam;

(************************************************)

type
 { TArrObj }

 TArrObj=class
   private
    type
    { TArrObjEnum }
     TArrObjEnum=class
        private
         fAO:TArrObj;
         index:integer;
        public
        constructor Create(AO:TArrObj);
        function GetCurrent:TObject;
        function MoveNext:boolean;
        procedure Reset;
        property Current: TObject read getCurrent;
     end;
    private
        fa: array of TObject;
        fown:boolean;
        function Getitem(index: integer): TObject;
        procedure Setitem(index: integer; AValue: TObject);
      public
        function GetEnumerator: TArrObjEnum;
      public
        constructor Create(own:boolean);
        destructor Destroy;override;
        function add(obj:TObject):TObject;
        function first:TObject;
        function last:TObject;
        function len:integer;
        property item[index:integer]:TObject read Getitem write Setitem; default;
        procedure clear;
 end;

implementation

uses SysUtils, typinfo;


var v4l:TWebCam=nil;
    pbuf:pword; // для динамического bw-тресхолд-анализа

function webcam:TWebCam; begin result:=v4l; end;

{ TCamImageBGRA }

function TCamImageBGRA.GetInternalColor(x, y: integer): TFPColor;
var
  v: TColorBGRA;
begin
  v:=PColorBGRA(FData)[x+y*Width];
  Result.red:=(v.r shl 8)+v.r;
  Result.green:=(v.g shl 8)+v.g;
  Result.blue:=(v.b shl 8)+v.b;
  Result.alpha:=(v.a shl 8)+v.a;
end;

procedure TCamImageBGRA.SetInternalColor(x, y: integer; const Value: TFPColor);
 var v: TColorBGRA;
begin
 v.r:=Value.red shr 8;
 v.g:=Value.green shr 8;
 v.b:=Value.blue shr 8;
 v.a:=Value.alpha shr 8;
 PColorBGRA(FData)[x+y*Width]:=v;
end;

constructor TCamImageBGRA.Create(AWidth, AHeight: integer);
begin
 FDesc:=GetFPCompactImgDesc(false,8,true);
 inherited Create(AWidth, AHeight);
end;

procedure TCamImageBGRA.SetSize(AWidth, AHeight: integer);
begin
  if (AWidth=Width) and (AHeight=Height) then exit;
  ReAllocMem(FData,SizeOf(TColorBGRA)*AWidth*AHeight);
  inherited SetSize(AWidth,AHeight);
end;

function TCamImageBGRA.YUYVImport(YUYV: PByte; _Width, _Height: integer
  ): TCamImage;
begin
  setsize(_width,_height);
  Twebcam.dec_yuyv_bgra(YUYV,data,_width,_height);
  result:=self;
end;

{ TCamImage }

function TCamImage.GetInternalPixel(x, y: integer): integer;
begin
 result:=0;
end;

procedure TCamImage.SetInternalPixel(x, y: integer; Value: integer);
begin
end;

destructor TCamImage.Destroy;
begin
  ReallocMem(fdata,0);
  inherited Destroy;
end;


{ TCamImageGray }

function TCamImageGray.GetInternalColor(x, y: integer): TFPColor;
begin
 Result.red:=FData[x+y*Width];
 Result.red:=(Word(Result.red) shl 8)+Result.red;
 Result.green:=Result.red;
 Result.blue:=Result.red;
 Result.alpha:=alphaOpaque;
end;

procedure TCamImageGray.SetInternalColor(x, y: integer; const Value: TFPColor);
begin
  FData[x+y*Width]:=Value.red shr 8;
end;

constructor TCamImageGray.Create(AWidth, AHeight: integer);
begin
  FDesc:=GetFPCompactImgDesc(true,8,false);
  inherited Create(AWidth, AHeight);
end;

procedure TCamImageGray.SetSize(AWidth, AHeight: integer);
begin
  if (AWidth=Width) and (AHeight=Height) then exit;
  ReAllocMem(FData,SizeOf(Byte)*AWidth*AHeight);
  inherited SetSize(AWidth, AHeight);
end;

function TCamImageGray.YUYVImport(YUYV: PByte; _Width, _Height: integer
  ): TCamImage;
begin
  setsize(_width,_height);
  Twebcam.dec_yuyv_gray(YUYV,data,_width,_height);
  result:=self;
end;

{ TCamImageRGB }

function TCamImageRGB.GetInternalColor(x, y: integer): TFPColor;
var
  v: TColorRGB;
begin
  v:=PColorRGB(FData)[x+y*Width];
  Result.red:=(v.r shl 8)+v.r;
  Result.green:=(v.g shl 8)+v.g;
  Result.blue:=(v.b shl 8)+v.b;
  Result.alpha:=alphaOpaque;
end;

procedure TCamImageRGB.SetInternalColor(x, y: integer; const Value: TFPColor);
 var
   v: TColorRGB;
 begin
   v.r:=Value.red shr 8;
   v.g:=Value.green shr 8;
   v.b:=Value.blue shr 8;
   PColorRGB(FData)[x+y*Width]:=v;
 end;

constructor TCamImageRGB.Create(AWidth, AHeight: integer);
begin
  FDesc:=GetFPCompactImgDesc(false,8,false);
  inherited Create(AWidth, AHeight);
end;

procedure TCamImageRGB.SetSize(AWidth, AHeight: integer);
begin
  if (AWidth=Width) and (AHeight=Height) then exit;
  ReAllocMem(FData,SizeOf(TColorRGB)*AWidth*AHeight);
  inherited SetSize(AWidth, AHeight);
end;

function TCamImageRGB.YUYVImport(YUYV: PByte; _Width, _Height: integer
  ): TCamImage;
begin
  setsize(_width,_height);
  Twebcam.dec_yuyv_rgb(YUYV,pbyte(data),_width,_height);
  result:=self;
end;



const eqError=0.0001;

function eq(left,right:double):boolean;
begin
  result := abs(left-right) < eqError;
end;

function le(left,right:double):boolean;
begin
  result := (left<=right) or eq(left,right);
end;

function ge(left,right:double):boolean;
begin
  result := (left>=right) or eq(left,right);
end;

function _FpOpen(path: PChar; flags: cInt): cInt; cdecl;
begin
  Result := FpOpen(path, flags);
end;

function _FpClose(fd: cInt): cInt; cdecl;
begin
  Result := FpClose(fd);
end;

function _FpIOCtl(Handle: cint; Ndx: TIOCtlRequest; Data: Pointer): cint; cdecl;
begin
  Result := FpIOCtl(Handle, Ndx, Data);
end;

function mmap(__addr:pointer; __len:size_t; __prot:longint; __flags:longint; __fd:longint;
           __offset:dword):pointer;cdecl;external 'c' name 'mmap';

function _Fpmmap(start: pointer; len: size_t; prot: cint; flags: cint; fd: cint; offst: off_t): pointer; cdecl;
begin
  Result := mmap(start, len, prot, flags, fd, offst);
end;

function _Fpmunmap(start: pointer; len: size_t): cint; cdecl;
begin
  Result := Fpmunmap(start, len);
end;

{ TPixelFormatDescription }

function TPixelFormatDescription.getDescription: string;
begin
  result:=pchar(@ffmt.description);
end;

function TPixelFormatDescription.getpixelFormat: TPixelFormat;
begin
 for result in TPixelFormat do begin
  if ConstsUVCPixelFormat[result]=V4L2_pixelFormat then exit;
 end;
 result:=pf_NONE;
end;

function TPixelFormatDescription.getpixelFormatName: string;
 var pname:pchar;
     i:integer;
begin
 result:='';
 pname:=@ffmt.pixelformat;
 for i:=0 to 3 do result+=pname[i];
end;

constructor TPixelFormatDescription.Create(_owner: TWebCam; fmt: v4l2_fmtdesc);
begin
  inherited create(_owner);
  ffmt:=fmt;
end;

{ TV4L2Control }

constructor TV4L2Control.Create(_owner: TWebCam);
begin
  inherited Create();
  fowner:=_owner;
end;

function TV4L2Control.valid: boolean;
begin
 result:=self<>nil;
end;

{ TFrameRate }

constructor TFrameRate.Create(_owner: TWebCam; framerate: v4l2_frmivalenum
  );
begin
  inherited create(_owner);
  fframerate:=framerate;
end;

procedure TFrameRate.setFrameRate();
begin
  owner.setFrameRate(width,height,denominator/numerator);
end;

{ TFrameControl }

function TFrameControl.getFrameRateCount: integer;
begin
  result:=fframerates.len;
end;

function TFrameControl.getFrameRate(index: integer): TFrameRate;
begin
 result:=fframerates[index]as TFrameRate;
end;

constructor TFrameControl.Create(_owner: TWebCam; frame: v4l2_frmsizeenum);
 var q2: v4l2_frmivalenum;
begin
  inherited create(_owner);
  fframeRates:=TArrObj.Create(true);
  fframe:=frame;
  q2.index:=0;
  q2.pixel_format:=fframe.pixel_format;
  q2.width:=fframe.u.discrete.width;
  q2.height:=fframe.u.discrete.height;
  while true do begin
     if _fpIoctl(fowner.Handle, VIDIOC_ENUM_FRAMEINTERVALS, @q2) <> 0 then break;
     fframerates.add(TFrameRate.Create(owner,q2));
     fowner.AddDebug(format(' framerate (type %d): [numerator %d, denominator %d]',
      [q2._type,q2.u.discrete.numerator,q2.u.discrete.denominator ] ));
     inc(q2.index);
  end;
end;

destructor TFrameControl.Destroy;
begin
  fframerates.free;
  inherited destroy;
end;

procedure TFrameControl.setFrameRate(Rate: double);
begin
  owner.setFrameRate(width,height,Rate);
end;

{ TUserControl }

function TUserControl.getName: string;
begin
  result:=fcontrol.name;
end;

function TUserControl.getValue: integer;
begin
  result:=fowner.getFeatureValue(fcontrol.id);
end;

procedure TUserControl.setValue(AValue: integer);
begin
 if not valid then exit;
 fowner.SetFeatureValue(fcontrol.id,Avalue);
end;

constructor TUserControl.Create(_owner: TWebCam; control: v4l2_queryctrl);
begin
  inherited create(_owner);
  fcontrol:=control;
end;

procedure TUserControl.setDefault;
begin
  value:=def;
end;

procedure TUserControl.setMin;
begin
 value:=min;
end;

procedure TUserControl.setMax;
begin
  value:=max;
end;


{ TWebCam }

procedure TWebCam.setFrameRate(_Width: integer; _Height: integer; _Rate: double);
begin
 if not fOpened then begin
    DeviceOpen(ParamNone,_Width,_Height,_Rate);
    exit;
 end;
 if FStarted then DeviceStop;
 GetParams(_Width, _Height, _Rate, fRateNum, fRateDenom);
 if (fwidth<>_Width)or(fheight<>_Height) then begin
  if assigned(fImageRGB) then fImageRGB.setsize(_width,_height)
  else fImageRGB:=TCamImageRGB.Create(fWidth,fHeight);
  if assigned(fImageBGRA) then fImageBGRA.setsize(_width,_height)
  else fImageBGRA:=TCamImageBGRA.Create(fWidth,fHeight);
  if assigned(fImageGray) then fImageGray.setsize(_width,_height)
  else fImageGray:=TCamImageGray.Create(fWidth,fHeight);
  if assigned(fImageYUYV) then reallocmem(fImageYUYV,_width*_height*2)
  else fImageYUYV:=allocmem(_width*_height*2);
 end;
 fWidth:=_Width;
 fHeight:=_Height;
end;



procedure TWebCam.AddDebug(s: string);
begin
  setlength(DebugList,length(DebugList)+1);
  DebugList[length(DebugList)-1]:=s;
end;


constructor TWebCam.Create(_prefixDevice: string);
begin
  inherited Create;
  _Frames:=TArrObj.Create(true);
  _ExtControls:=TArrObj.Create(true);
  _PixelFormats:=TArrObj.Create(true);
  FHandle := -1;
  DebugList := nil;
  fDevice:='';
  fPrefixDevice:=_prefixDevice;
end;

destructor TWebCam.Destroy;
begin
  DeviceStop;
  DeviceClose;
  _ExtControls.free;
  _Frames.free;
  _PixelFormats.free;
  fImageRGB.free;
  fImageBGRA.free;
  fImageGray.free;
  reallocMem(fImageYUYV,0);
  inherited Destroy;
end;


procedure TWebCam.GetUserControls;
var queryctrl: v4l2_queryctrl;
    id, i: integer;
    control: TUserControl;
begin
  if FHandle = -1 then exit;
  for i:=low(TV4L2_CIDS) to high(TV4L2_CIDS) do FreeAndNil(_UserControls[i]);

  fillbyte(queryctrl, sizeof(queryctrl), 0);
  queryctrl.id := V4L2_CTRL_FLAG_NEXT_CTRL;
  while _fpIoctl(FHandle, VIDIOC_QUERYCTRL, @queryctrl) = 0 do begin
    control:=TUserControl.Create(self,queryctrl);
    if (queryctrl.id>=low(TV4L2_CIDS))and (queryctrl.id<=high(TV4L2_CIDS))
     then _UserControls[queryctrl.id] := control
     else _ExtControls.add(control);

     AddDebug(pchar(@queryctrl.name[0]) + format('; id: %d (type %d): [min %d, max %d, step %d, def %d, current %d]',
      [queryctrl.id, queryctrl._type, queryctrl.minimum, queryctrl.maximum, 
       queryctrl.step, queryctrl.default_value, GetFeatureValue(queryctrl.id) ] ));

    id := queryctrl.id or V4L2_CTRL_FLAG_NEXT_CTRL;
    fillbyte(queryctrl, sizeof(queryctrl), 0);
    queryctrl.id := id;
  end;

end;

procedure TWebCam.GetFrameControls;
 var queryctrl: v4l2_frmsizeenum;
begin
  if FHandle = -1 then exit;
  queryctrl.index:=0;
  queryctrl.pixel_format:=ConstsUVCPixelFormat[fPixelFormat];
  while true do begin
      if fpIoctl(Handle, VIDIOC_ENUM_FRAMESIZES, @queryctrl) <> 0 then break;
      AddDebug('Support FrameSize: '+format('(type %d): [width %d, height %d]',
       [queryctrl._type,queryctrl.u.discrete.width,queryctrl.u.discrete.height]));
      _Frames.add(TFrameControl.Create(self,queryctrl));
      inc(queryctrl.index);
  end;
end;

function TWebCam.getV4L2_PixelFormat: longint;
begin
 result:=ConstsUVCPixelFormat[fPixelFormat];
end;

function TWebCam.CapabilitiesToCamCaps(v4l2caps: __u32): TCamCaps;
  var caps:__u32 absolute result;
begin
 caps:=v4l2caps;
end;

var
 // описание возможностей камеры
 _CamCapInfo:array[TCamCap]of ansistring=(
 'V4L2_CAP_VIDEO_CAPTURE: The device supports the single-planar API through the Video Capture interface.',
 'V4L2_CAP_VIDEO_OUTPUT: The device supports the single-planar API through the Video Output interface.',
 'V4L2_CAP_VIDEO_OVERLAY: The device supports the Video Overlay interface. A video overlay device typically stores captured images directly in the video memory of a graphics card, with hardware clipping and scaling.',
 '',
 'V4L2_CAP_VBI_CAPTURE: The device supports the Raw VBI Capture interface, providing Teletext and Closed Caption data.',
 'V4L2_CAP_VBI_OUTPUT: The device supports the Raw VBI Output interface.',
 'V4L2_CAP_SLICED_VBI_CAPTURE: The device supports the Sliced VBI Capture interface.',
 'V4L2_CAP_SLICED_VBI_OUTPUT: The device supports the Sliced VBI Output interface.',
 'V4L2_CAP_RDS_CAPTURE: The device supports the RDS capture interface.',
 'V4L2_CAP_VIDEO_OUTPUT_OVERLAY: The device supports the Video Output Overlay (OSD) interface. Unlike the Video Overlay interface, this is a secondary function of video output devices and overlays an image onto an outgoing video signal. When the driver sets this flag, it must clear the V4L2_CAP_VIDEO_OVERLAY flag and vice versa.',
 'V4L2_CAP_HW_FREQ_SEEK: The device supports the ioctl VIDIOC_S_HW_FREQ_SEEK ioctl for hardware frequency seeking.',
 'V4L2_CAP_RDS_OUTPUT: The device supports the RDS output interface.',
 'V4L2_CAP_VIDEO_CAPTURE_MPLANE: The device supports the multi-planar API through the Video Capture interface.',
 'V4L2_CAP_VIDEO_OUTPUT_MPLANE: The device supports the multi-planar API through the Video Output interface.',
 'V4L2_CAP_VIDEO_M2M: The device supports the single-planar API through the Video Memory-To-Memory interface.',
 'V4L2_CAP_VIDEO_M2M_MPLANE: The device supports the multi-planar API through the Video Memory-To-Memory interface.',
 'V4L2_CAP_TUNER: The device has some sort of tuner to receive RF-modulated video signals. For more information about tuner programming see Tuners and Modulators.',
 'V4L2_CAP_AUDIO: The device has audio inputs or outputs. It may or may not support audio recording or playback, in PCM or compressed formats. PCM audio support must be implemented as ALSA or OSS interface. For more information on audio inputs and outputs see Audio Inputs and Outputs.',
 'V4L2_CAP_RADIO: This is a radio receiver.',
 'V4L2_CAP_MODULATOR: The device has some sort of modulator to emit RF-modulated video/audio signals. For more information about modulator programming see Tuners and Modulators.',
 'V4L2_CAP_SDR_CAPTURE: The device supports the SDR Capture interface.',
 'V4L2_CAP_EXT_PIX_FORMAT: The device supports the struct v4l2_pix_format extended fields.',
 'V4L2_CAP_SDR_OUTPUT: The device supports the SDR Output interface.',
 'V4L2_CAP_META_CAPTURE: The device supports the Metadata Interface capture interface.',
 'V4L2_CAP_READWRITE: The device supports the read() and/or write() I/O methods.',
 'V4L2_CAP_ASYNCIO: The device supports the asynchronous I/O methods.',
 'V4L2_CAP_STREAMING: The device supports the streaming I/O method.',
 'V4L2_CAP_CAP_META_OUTPUT: The device supports the Metadata Interface output interface.',
 'V4L2_CAP_TOUCH: This is a touch device.',
 '',
 '',
 'V4L2_CAP_DEVICE_CAPS: The driver fills the device_caps field. This capability can only appear in the capabilities field and never in the device_caps field.'
 );

function TWebCam.getCamCapInfo(index: TCamCap): ansistring;
begin
 result:=_CamCapInfo[index];
end;

function TWebCam.GetFrameControl(index: integer): TFrameControl;
begin
  result:=_Frames[index] as TFrameControl;
end;

function TWebCam.GetFramesCount: integer;
begin
 result:=_Frames.len;
end;

function TWebCam.GetExtControl(index: integer): TUserControl;
begin
  result:=_extControls[index] as TUserControl;
end;

function TWebCam.GetExtControlsCount: integer;
begin
 result:=_extControls.len;
end;


procedure TWebCam.GetPixelFormats;
 var fmt:  v4l2_fmtdesc;
     index: dword;
     pf:TPixelFormatDescription;
begin
 _PixelFormats.Clear;
 index:=0;
 while true do begin
  fillchar(fmt,sizeof(fmt),0);
  fmt.index := index;
  fmt._type := V4L2_BUF_TYPE_VIDEO_CAPTURE;
  if _fpioctl(fHandle, VIDIOC_ENUM_FMT, @fmt) = -1 then break;
  pf:=TPixelFormatDescription.Create(self,fmt);
  _PixelFormats.add(pf);
  AddDebug('Support PixelFormat: '+format('ID: %d; Name: %s; Description: %s',
       [pf.V4L2_pixelFormat,pf.Name,pf.description]));
  inc(index);
 end;
end;

function TWebCam.GetUserControl(index: integer): TUserControl;
begin
 result:=_UserControls[index+V4L2_CID_BASE];
end;

function TWebCam.GetUserControlsCount: integer;
begin
 result:=length(_UserControls);
end;


function TWebCam.DeviceOpen(
                                  _Device:integer = ParamMin;
                                  _Width:integer = ParamMax;
                                  _Height:integer = ParamMax;
                                  _Rate:double = ParamMax;
                                  _pixelFormat:TPixelFormat=pf_YUYV):boolean;
var num:integer;
begin
 result:=true;
 try
  if FHandle<>-1 then DeviceClose;
  if _Device in [0..9] then fdevice:='';
  if (_device=ParamNone)and(fdevice='') then _device:=paramMin;
  if (_device=ParamMin)or(_device=ParamMax)  then begin
    num:=0;
    while not fileExists(FPrefixDevice+IntToStr(num)) do begin
      inc(num);
      if num>9 then raise Exception.Create('video device not found!');
    end;
    fdevice:=FPrefixDevice+IntToStr(num);
  end;
  if (_device=ParamMax) then begin
    while fileExists(FPrefixDevice+IntToStr(num)) do inc(num);
    fdevice:=FPrefixDevice+IntToStr(num-1);
  end;

  if fdevice='' then fdevice:=FPrefixDevice+IntToStr(_Device);

  FHandle := _fpOpen(pchar(FDevice), O_RDWR);
  if FHandle = -1 then
    raise Exception.create('could not open video device ' + FDevice);

  GetUserControls;
  GetPixelFormats;
  if _pixelFormat=pf_NONE then _pixelFormat:=pf_YUYV;

  fPixelFormat:=_pixelFormat;
  GetFrameControls;

  fOpened:=true; // обязательно ДО setFrameRate, иначе - бесконечный цикл
  setFrameRate(_Width, _Height, _Rate);

  fillbyte(fCap, sizeof(fCap), 0);
  _fpIoctl(FHandle, VIDIOC_QUERYCAP, @fCap);
  fDriver:=pchar(@fcap.driver);
  fCard:=pchar(@fcap.card);
  fBusInfo:=pchar(@fcap.bus_info);
  fCamCaps:=CapabilitiesToCamCaps(fcap.device_caps);

  addDebug(format( 'Setting params:'#10' Device=%s;'#10' Width=%d;'#10' Height=%d;'#10' Rate=%d/%d;'#10' PixelFormat=%s',
                   [fDevice,fwidth,fheight,frateNum,fRateDenom,getEnumName(typeinfo(TPixelFormat),ord(fpixelformat))]));
 except
   on e:exception do begin
     LastError:=e.message;
     result:=false;
   end;
 end;
end;

function TWebCam.DeviceStart:boolean;
var fmt: v4l2_format;
    parms: v4l2_streamparm;
    reqbuf: v4l2_requestbuffers;
    vbuf: v4l2_buffer;
    i: integer;
begin
   try
    if FHandle = -1 then deviceOpen();
    if fStarted then DeviceStop();

    // set webcam format
    FillChar(fmt, sizeof(v4l2_format), 0);
    fmt._type := V4L2_BUF_TYPE_VIDEO_CAPTURE;
    fmt.fmt.pix.width := FWidth;
    fmt.fmt.pix.height := FHeight;

    fmt.fmt.pix.pixelformat := ConstsUVCPixelFormat[FPixelFormat];
    //writeln (FPixelFormat);
    fmt.fmt.pix.field := V4L2_FIELD_ANY;
    if _fpIoctl(FHandle, VIDIOC_S_FMT, @fmt) < 0 then
      raise Exception.Create('could not set video format');

    // set the proterty values to what the driver was able to provide
    FWidth := fmt.fmt.pix.width;
    FHeight := fmt.fmt.pix.height;

    // set frame rate
    FillChar(parms, sizeof(v4l2_streamparm), 0);
    parms._type := V4L2_BUF_TYPE_VIDEO_CAPTURE;
    parms.parm.capture.timeperframe.numerator := fRateNum;
    parms.parm.capture.timeperframe.denominator := FRateDenom;
    if _fpIoctl(FHandle, VIDIOC_S_PARM, @parms) = -1 then
      raise Exception.Create('could not set desired frame rate');

    // set up buffers
    FillChar(reqbuf, sizeof(v4l2_requestbuffers), 0);
    reqbuf.count := VIDEO_BUFFERS;
    reqbuf._type := V4L2_BUF_TYPE_VIDEO_CAPTURE;
    reqbuf.memory := V4L2_MEMORY_MMAP;
    if _fpIoctl(FHandle, VIDIOC_REQBUFS, @reqbuf) < 0 then
      raise Exception.Create('could not set up video buffers');

    // mmap the buffers into process memory
    for i := 0 to VIDEO_BUFFERS - 1 do begin
      FillChar(vbuf, sizeof(v4l2_buffer), 0);
      vbuf.index := i;
      vbuf._type := V4L2_BUF_TYPE_VIDEO_CAPTURE;
      vbuf.memory := V4L2_MEMORY_MMAP;
      if _fpIoctl(FHandle, VIDIOC_QUERYBUF, @vbuf) < 0 then
        raise Exception.Create('Could not query video buffer');

      VideoBufferLength:=vbuf.length;
      VideoBuffer[i] := _fpMmap(nil, vbuf.length, PROT_READ, MAP_SHARED, FHandle, vbuf.m.offset);
      if VideoBuffer[i] = MAP_FAILED then
        raise Exception.Create('Could not mmap video buffer');
    end;

    // queue the buffers
    for i := 0 to VIDEO_BUFFERS - 1 do begin
      FillChar(vbuf, sizeof(v4l2_buffer), 0);
      vbuf.index := i;
      vbuf._type := V4L2_BUF_TYPE_VIDEO_CAPTURE;
      vbuf.memory := V4L2_MEMORY_MMAP;
      if _fpIoctl(FHandle, VIDIOC_QBUF, @vbuf) < 0 then
        raise Exception.Create('Could not queue video buffer');
    end;

    i := V4L2_BUF_TYPE_VIDEO_CAPTURE;
    if _fpIoctl(FHandle, VIDIOC_STREAMON, @i) < 0 then
      raise Exception.Create('Could not start streaming');

    fStarted:=true;
   except
     on e:exception do begin
       LastError:=e.message;
       result:=false;
     end;
   end;
end;

function TWebCam.Capture(count:integer=1):boolean;
begin
 result:=self.capture(TWebCamFrameEventProc(nil),count);
end;

function TWebCam.Capture(proc: TWebCamFrameEventNestedProc; count: integer):boolean;
    var vbuf: v4l2_buffer;
begin
   result:=true;
   if not fStarted then deviceStart();
   CurCaptureFrame := 0;
   while count<>0 do begin
    FillChar(vbuf, sizeof(v4l2_buffer), 0);
    vbuf._type := V4L2_BUF_TYPE_VIDEO_CAPTURE;
    vbuf.memory := V4L2_MEMORY_MMAP;
    if _fpIoctl(FHandle, VIDIOC_DQBUF, @vbuf) < 0 then
     begin
        lasterror:=format('Error prepare buffer (VIDIOC_DQBUF): %d',[fpGetErrno()]);
        result:=false;
        break;
     end;

    if vbuf.bytesused > 0 then begin
      VideoBufferIndex := vbuf.index;
      CurFramePtr:=VideoBuffer[VideoBufferIndex];
      if assigned(proc) then
        begin if not proc(self,VideoBuffer[VideoBufferIndex],CurCaptureFrame) then count:=1;end
      else
      if Assigned(FOnFrame) then begin
        if not FOnFrame(self,VideoBuffer[VideoBufferIndex],CurCaptureFrame) then count:=1;
      end;
    end;

    if _fpIoctl(FHandle, VIDIOC_QBUF, @vbuf) < 0 then begin
      lasterror:=format('Error resubmit buffer (VIDIOC_QBUF): %d',[fpGetErrno()]);
      result:=false;
      break;
    end;
    Inc(CurCaptureFrame);
    dec(count);
   end;
end;

function TWebCam.Capture(proc: TWebCamFrameEventProc; count: integer):boolean;
   function _proc(Sender: TWebCam; FramePtr: PByte; FrameNumber:integer):boolean;
    begin result:=proc(Sender,FramePtr,FrameNumber); end;
begin
  if assigned(proc) then
   result:=capture(@_proc,count)
  else
   result:=capture(TWebCamFrameEventNestedProc(nil),count);
end;

function TWebCam.CaptureImageRGB(skip: integer): TCamImageRGB;
  function RGBImageFrame(Sender: TWebCam; FramePtr: PByte; FrameNumber:integer):boolean;
   begin
     result:=true;
     if framenumber=skip then begin
      fImageRGB.YUYVImport(FramePtr,Width,Height);
     end;
   end;

begin
  if not assigned(fImageRGB) then fImageRGB:=TCamImageRGB.Create(Width,Height);
  result:=fImageRGB;
  capture(@RGBImageFrame,skip+1);
end;

function TWebCam.CaptureImageBGRA(skip: integer): TCamImageBGRA;
 function BGRAImageFrame(Sender: TWebCam; FramePtr: PByte; FrameNumber:integer):boolean;
  begin
   result:=true;
   if framenumber=skip then begin
    fImageBGRA.YUYVImport(FramePtr,Width,Height);
   end;
 end;

begin
 if not assigned(fImageBGRA) then fImageBGRA:=TCamImageBGRA.Create(Width,Height);
 result:=fImageBGRA;
 capture(@BGRAImageFrame,skip+1);
end;

function TWebCam.CaptureImageGray(skip: integer): TCamImageGray;
  function GrayImageFrame(Sender: TWebCam; FramePtr: PByte; FrameNumber:integer):boolean;
   begin
     result:=true;
     if framenumber=skip then begin
      fImageGray.YUYVImport(FramePtr,Width,Height);
     end;
   end;

begin
  if not assigned(fImageGray) then fImageGray:=TCamImageGray.Create(Width,Height);
  result:=fImageGray;
  capture(@GrayImageFrame,skip+1);
end;

function TWebCam.CaptureImageYUYV(skip: integer): PByte;
function YUYVImageFrame(Sender: TWebCam; FramePtr: PByte; FrameNumber:integer):boolean;
 begin
   result:=true;
   if framenumber=skip then begin
    move(FramePtr^,fImageYUYV^,Width*Height*2);
   end;
 end;

begin
 if not assigned(fImageYUYV) then begin
    fImageYUYV:=AllocMem(Width*Height*2);
    fImageYUYVWidth:=Width; fImageYUYVHeight:=Height;
 end
 else
 if (fImageYUYVWidth<>Width)or(fImageYUYVHeight<>Height) then begin
   ReAllocMem(fImageYUYV,Width*Height*2);
   fImageYUYVWidth:=Width; fImageYUYVHeight:=Height;
 end;
 result:=fImageYUYV;
 capture(@YUYVImageFrame,skip+1);
end;



procedure TWebCam.DeviceClose;
begin
  ClearDebugList();
  if FHandle = -1 then exit;
  DeviceStop;
  _fpClose(FHandle);
  FHandle := -1;
  fOpened := false;
//  _fDevice:='';
end;


procedure TWebCam.getParams(var _Width: integer; var _Height: integer;
   _Rate:double; out _RateNum:integer; out _RateDenom:integer);
  var f,found:TFrameControl;
      fr,foundFR:TFrameRate;
begin
 if fHandle=-1 then exit;
 if _Width=ParamNone then _Width:=fWidth;
 if _Height=ParamNone then _Height:=fHeight;
 if eq(_Rate,ParamNone) then begin
  _RateNum:=fRateNum;
  _ratedenom:=fRateDenom;
 end;

 found:=_Frames[0] as tframeControl;
 if _Width=ParamMax then begin
   for TObject(f) in _Frames do begin
     if f.width>found.width then found:=f;
   end;
   _Width:=found.width;
   getParams(_Width,_Height,_Rate, _RateNum, _RateDenom);//,found);
 end else
 if _Width=ParamMin then begin
   for TObject(f) in _Frames do begin
     if f.width<found.width then found:=f;
   end;
   _Width:=found.width;
   getParams(_Width,_Height,_Rate, _RateNum, _RateDenom);//,found);
 end;

 found:=_Frames[0] as tframeControl;
 for TObject(f) in _Frames do
   if f.width=_Width then found:=f;

 if (_Height=paramMax) then begin
    for TObject(f) in _Frames do begin
      if (f.width=_Width)and(f.height>=found.height) then begin
      found:=f;
     end;
    end;
    _Height:=found.height;
    getParams(_Width,_Height,_Rate, _RateNum, _RateDenom);//,found);
 end else
 if (_Height=paramMin) then begin
     for TObject(f) in _Frames do begin
      if (f.width=_Width)and(f.height<=found.height) then found:=f;
     end;
     _Height:=found.height;
     getParams(_Width,_Height,_Rate, _RateNum, _RateDenom);//,found);
 end;


 found:=_Frames[0] as tframeControl;
 for TObject(f) in _Frames do begin
     if (f.width=_Width)and(f.height=_Height) then found:=f;
 end;
 foundFR:=found.fframerates[0] as TFrameRate;

 if (eq(_Rate,paramMax)) then begin
    for TObject(fr) in found.fframerates do begin
       if (fr.denominator/fr.numerator) > (foundfr.denominator/foundfr.numerator) then foundFR:=fr;
    end;
 end else
 if (eq(_Rate,paramMin)) then begin
    for TObject(fr) in found.fframerates do begin
       if (fr.denominator/fr.numerator) < (foundfr.denominator/foundfr.numerator) then foundFR:=fr;
    end;
 end
 else begin
   for TObject(fr) in found.fframerates do begin
      if eq((fr.denominator/fr.numerator),_Rate) then begin
        foundFR:=fr; break;
      end;
   end;
 end;

 _Width:=found.width;
 _Height:=found.height;
 _Rate:=foundFR.denominator/foundfr.numerator;
 _RateNum:=foundfr.numerator;
 _RateDenom:=foundfr.denominator;
// writeln(_width,' ',_height,' ',_RateDenom,' ',_RateNum);
end;



function TWebCam.getRate: double;
begin
  result:=fRateDenom/fRateNum;
end;


procedure TWebCam.ClearDebugList();
begin
  setlength(DebugList,0);
end;


procedure TWebCam.DeviceStop;
var i,value:integer;
begin
  if fHandle=-1 then exit;
  if not fStarted then exit;
  for i:=low(TV4L2_CIDS) to high(TV4L2_CIDS) do FreeAndNil(_UserControls[i]);
  _ExtControls.clear;
  _Frames.clear;
  value := V4L2_BUF_TYPE_VIDEO_CAPTURE;
  _fpIoctl(FHandle, VIDIOC_STREAMOFF, @value);
  for i := 0 to VIDEO_BUFFERS - 1 do begin
      if VideoBuffer[i] <> MAP_FAILED then
        _fpMunmap(VideoBuffer[i],VideoBufferLength);
  end;
  fStarted:=false;
end;

procedure TWebCam.setDefault;
begin
 if Brightness.valid then Brightness.setDefault;
 if Contrast.valid then Contrast.setDefault;
 if Saturation.valid then Saturation.setDefault;
 if Hue.valid then Hue.setDefault;
 if Gamma.valid then Gamma.setDefault;
 if Sharpness.valid then Sharpness.setDefault;
 if BacklightCompensation.valid then BacklightCompensation.setDefault;
end;

function TWebCam.GetFeatureValue(feature: LongWord): integer;
var control_s: v4l2_control;
begin
  result := 0;
  if FHandle = -1 then exit;
  control_s.id := feature;
  if _fpIoctl(FHandle, VIDIOC_G_CTRL, @control_s) < 0 then
    raise Exception.Create('Could not get feature '+ inttohex(feature, 8));
  result := control_s.value;
end;

procedure TWebCam.SetFeatureValue(feature: LongWord; value: integer);
var control_s: v4l2_control;
begin
  if FHandle = -1 then exit;
  control_s.id := feature;
  control_s.value := value;
  if _fpIoctl(FHandle, VIDIOC_S_CTRL, @control_s) < 0 then
    raise Exception.Create('Could not set feature '+ inttohex(feature, 8) + ' to value ' + inttostr(value));
end;

function TWebCam.TrySetFeatureValue(feature: LongWord; value: integer): boolean;
var control_s: v4l2_control;
begin
  result := false;
  if FHandle = -1 then exit;
  control_s.id := feature;
  control_s.value := value;
  if not (_fpIoctl(FHandle, VIDIOC_S_CTRL, @control_s) < 0) then
    result := true;
end;



class function TWebCam.dec_yuyv_bw(src:pbyte; dest:pbyte; width,height:integer; bound:byte):boolean;
   var i:integer;
  begin
   result:=true;
   for i := 0 to (width*height div 2) - 1 do begin
    if src[0]>bound then dest[0]:=255 else dest[0]:=0;
    if src[2]>bound then dest[1]:=255 else dest[1]:=0;
    Inc(dest,2);
    Inc(src,4);
   end;
  end;



class function TWebCam.dec_yuyv_bwdynamic(src: pbyte; dest: pbyte; width,
  height: integer; divX, divY, c: integer): boolean;
var
      color:word;
      pos,linepos,colpos:integer;
      ih,iw,W,H:integer;
      gbound:dword;
  const mulK=2; // мультипликационный коэффициент, для уменьшения потерь при делении
                // дает возможность работать с областями 51x51 точек
begin
 result:=true;
// двумерный массив средних величин по горизонтали
 ReAllocMem(pbuf,width*height*2);
 // радиус сегмента по горизонтали и вертикали
 W:=width div divX; H:=height div divY;
// writeln (w, ' ', h);
 // заполнение строк суммами в буфере pbuf
 for ih:=0 to height-1 do begin
  // заполнение первой ячейки строки
  color:=0;
  linepos:=ih*width;
  pos:=linepos;
  for iw:=0 to W do begin
   inc(color,src[pos*2]);
   inc(pos);
  end;
  pos:=linepos;
  pbuf[pos]:=(color shl mulk) div (w+1);
  // заполнение начальных ячеек суммированием без вычетов
  for iw:=1 to W do begin
   inc(pos);
   inc(color,src[(pos+w)*2]);
   pbuf[pos]:=(color shl mulk) div (w+1+iw);
  end;
  // заполнение центральных ячеек суммированием и вычетами
  for iw:=W+1 to (width-1-W) do begin
   inc(pos);
   inc(color,src[(pos+w)*2]);
   dec(color,src[(pos-w-1)*2]);
   pbuf[pos]:=(color shl mulk) div (2*w+1);
  end;
  // заполнение конечных ячеек вычетами без суммирования
  for iw:=W-1 downto 0 do begin
   inc(pos);
   dec(color,src[(pos-w-1)*2]);
   pbuf[pos]:=(color shl mulk) div (w+1+iw);
  end;
  // строка заполнена суммами
 end;

 // заполнение целевого изображения средними величинами из буфера pbuf
 for iw:=0 to width-1 do begin
  // заполнение первой ячейки столбца
  color:=0;
  colpos:=iw;
  pos:=colpos;
  for ih:=0 to H do begin
   inc(color,pbuf[pos]);
   inc(pos,width);
  end;
  pos:=colpos;
  dest[pos]:=(color shr mulk) div (h+1);
  // заполнение начальных ячеек суммированием без вычетов
  for ih:=1 to H do begin
   inc(pos,width);
   inc(color,pbuf[pos+H*width]);
   dest[pos]:=(color shr mulk) div (H+1+ih);
  end;
  // заполнение центральных ячеек суммированием и вычетами
  for ih:=h+1 to (height-1-h) do begin
   inc(pos,width);
   inc(color,pbuf[pos+h*width]);
   dec(color,pbuf[pos-(h+1)*width]);
   dest[pos]:=(color shr mulk) div (2*h+1);
  end;
  // заполнение конечных ячеек вычетами без суммирования
  for ih:=H-1 downto 0 do begin
   inc(pos,width);
   dec(color,pbuf[pos-(h+1)*width]);
   dest[pos]:=(color shr mulk) div (h+1+ih);
  end;
 end;

 // среднее
 gbound:=0;
 for ih:=0 to height-1 do begin
  pos:=ih*width;
  for iw:=0 to width-1 do begin
   inc(gbound,src[pos*2]);
   inc(pos);
  end;
 end;
 gbound:=gbound div (width*height);

for ih:=0 to height-1 do begin
 pos:=ih*width;
 for iw:=0 to width-1 do begin
  if src[pos*2]>dest[pos]-c
    then dest[pos]:=255 else dest[pos]:=0;
  inc(pos);
 end;
end;

end;



class function TWebCam.dec_gray_bgra(src: pbyte; dest: pbyte; width,
  height: integer): boolean;
var i:integer;
begin
 result:=true;
 for i := 0 to width*height - 1 do begin
  dest[0] := src[0];
  dest[1] := src[0];
  dest[2] := src[0];
  dest[3] := 0;
  Inc(dest,4);
  Inc(src);
 end;
end;




class function TWebCam.dec_yuyv_gray(src:pbyte; dest:pbyte; width,height:integer):boolean;
    var i:integer;
  begin
   result:=true;
   for i := 0 to (width*height div 2) - 1 do begin
    dest[0] := src[0];
    dest[1] := src[2];
    Inc(dest,2);
    Inc(src,4);
   end;
  end;




class function TWebCam.dec_yuyv_rgb(src:pbyte; dest:pbyte; width,height:integer; pal:integer=PAL_YUYV):boolean;
   var 	r,g,b:integer;
        y,u,v:integer;
	z:boolean=false;
        i,j:integer;
  begin
   result:=true;
   for i:=1 to height do begin
    for j:=1 to width do begin
     if pal=PAL_UYVY then begin
       if(not z)then y:=src[1] shl 8 else y:=src[3] shl 8;
       u:= src[0]-128; v:= src[2]-128;
     end else
     if pal=PAL_YUYV then begin
       if(not z) then y:=src[0] shl 8 else y:=src[2] shl 8;
       u:= src[1]-128; v:= src[3]-128;
     end else
     if pal=PAL_VYUY then begin
       if(not z) then y:=src[1] shl 8 else y:=src[3] shl 8;
       u:= src[2]-128; v:= src[0]-128;
     end;

     r:= (y + (359 * v)) shr 8;
     g:= (y - (88 * u) - (183 * v)) shr 8;
     b:= (y + (454 * u)) shr 8;

     if r>255 then r:=255 else if r<0 then r:=0;
     if g>255 then g:=255 else if g<0 then g:=0;
     if b>255 then b:=255 else if b<0 then b:=0;

     dest[0]:=r;
     dest[1]:=g;
     dest[2]:=b;
     inc(dest,3);

     if z then inc(src,4);
     z:=not z;
    end
   end;
  end;

class function TWebCam.dec_yuyv_bgra(src: pbyte; dest: pbyte; width,
  height: integer; pal: integer): boolean;
var 	r,g,b:integer;
     y,u,v:integer;
     z:boolean=false;
     i,j:integer;
begin
result:=true;
for i:=1 to height do begin
 for j:=1 to width do begin
  if pal=PAL_UYVY then begin
    if(not z)then y:=src[1] shl 8 else y:=src[3] shl 8;
    u:= src[0]-128; v:= src[2]-128;
  end else
  if pal=PAL_YUYV then begin
    if(not z) then y:=src[0] shl 8 else y:=src[2] shl 8;
    u:= src[1]-128; v:= src[3]-128;
  end else
  if pal=PAL_VYUY then begin
    if(not z) then y:=src[1] shl 8 else y:=src[3] shl 8;
    u:= src[2]-128; v:= src[0]-128;
  end;

  r:= (y + (359 * v)) shr 8;
  g:= (y - (88 * u) - (183 * v)) shr 8;
  b:= (y + (454 * u)) shr 8;

  if r>255 then r:=255 else if r<0 then r:=0;
  if g>255 then g:=255 else if g<0 then g:=0;
  if b>255 then b:=255 else if b<0 then b:=0;

  dest[0]:=b;
  dest[1]:=g;
  dest[2]:=r;
  dest[3]:=0;
  inc(dest,4);

  if z then inc(src,4);
  z:=not z;
 end
end;
end;


{ TArrObjEnum }

constructor TArrObj.TArrObjEnum.Create(AO: TArrObj);
begin
  inherited create;
  fAO:=AO;
  reset;
end;

function TArrObj.TArrObjEnum.GetCurrent: TObject;
begin
   result:=fAO.fa[index];
end;

function TArrObj.TArrObjEnum.MoveNext: boolean;
begin
 if index>=length(fAO.fa)-1 then exit(false);
 inc(index);
 result:=true;
end;

procedure TArrObj.TArrObjEnum.Reset;
begin
 index:=-1;
end;

{ TArrObj }

function TArrObj.Getitem(index: integer): TObject;
begin
 if (index<0)or(index>=length(fa)) then exit(nil);
 result:=fa[index];
end;

procedure TArrObj.Setitem(index: integer; AValue: TObject);
begin
 if (index<0) then exit;
 if (index>=length(fa)) then setlength(fa,index-1);
 fa[index]:=AValue;
end;

constructor TArrObj.Create(own: boolean);
begin
  inherited create;
  fown:=own;
end;

destructor TArrObj.Destroy;
begin
  clear;
  inherited Destroy;
end;

function TArrObj.add(obj: TObject): TObject;
begin
 result:=obj;
 setlength(fa,length(fa)+1);
 fa[length(fa)-1]:=obj;
end;

function TArrObj.first: TObject;
begin
  result:=fa[0];
end;

function TArrObj.last: TObject;
begin
  result:=fa[length(fa)-1];
end;

function TArrObj.len: integer;
begin
 result:=length(fa);
end;

procedure TArrObj.clear;
 var
   v: TObject;
 begin
   if fown then for v in fa do v.free;
   setLength(fa,0);
end;

function TArrObj.GetEnumerator: TArrObjEnum;
begin
    result:=TArrObjEnum.Create(self);
end;



initialization
  v4l:=TWebCam.Create;
finalization
  v4l.Free;
  ReAllocMem(pbuf,0);
end.
